from __future__ import absolute_import, unicode_literals, print_function

import os
import re
import shutil
import subprocess
import sys

from select import select as pyselect
from config_installer.parser import ExtendedConfigParser


str_type = str
if sys.version_info[0] < 3:
    str_type = unicode  # noqa
    FileExistsError = OSError


class InstallationError(Exception):
    pass


class NotInstalled(Exception):
    pass


class BaseConfigInstaller(object):
    """
    Base class to install a configuration file.

    Subclass to create your installer:

    >>> from config_installer import BaseConfigInstaller
    >>> class MyInstaller(BaseConfigInstaller):
    ...     project_name = 'myproject'
    ...     config_basename = 'myproject.ini'
    ...
    >>> installer = MyInstaller()
    >>> config = installer.load()
    """

    #: :class:`str` - Name of your project.
    project_name = None

    #: Basename of the main configuration file to check for existence.
    #: Set to ``None`` to check the directory instead. This is the default.
    config_basename = None

    #: :class:`dict` - A dictionary to set options in the DEFAULT section
    config_defaults = {}

    #: :class:`bool` - include /etc/<project_name>/
    enable_etc = False

    #: :class:`bool` - include xdg-standard locations
    enable_xdg = False

    #: :class:`list` - List of additional directories
    additional_dirs = []

    #: :class:`list` - List of additional hidden directories. These are not
    #: shown in the :meth:`ask_for_dir` selection.
    additional_hidden_dirs = []

    #: The mode to set on a created installation directory
    create_mode = 0o750

    #: Set to ``True`` to replace a ``;config dir =`` line in the created
    #: configuration file.
    set_configdir = False

    def __init__(self):
        self.project_name_upper = self.project_name.upper()
        self.is_root = os.geteuid() == 0

        self.config = self.get_config_parser()
        defaults = self.get_config_defaults()
        for key, value in defaults.items():
            self.config.set('DEFAULT', key, self.format_default_value(value))

        self.loaded_config_paths = []
        self.load()

        self.install_dir = None

    def format_default_value(self, value):
        """Format python value for configparser DEFAULT section."""
        if isinstance(value, (list, tuple)):
            value = '\n    ' + ',\n    '.join([str_type(v) for v in value])
        return str_type(value)

    def get_xdg_dirs(self):
        """Yield available configuration file directories."""
        if self.is_root:
            dirs = os.environ.get('XDG_CONFIG_DIRS', '/etc/xdg').split(':')
            for path in dirs:
                path = path.strip()
                if not path:
                    continue
                yield path

        yield os.environ.get('XDG_CONFIG_HOME',
                             os.path.join(os.path.expanduser('~'), '.config'))

    def get_dirs(self, all=False):
        """Yield available configuration directory paths."""
        dirs = []

        if all or (self.enable_etc and self.is_root):
            dirs.append('/etc')

        if all or self.enable_xdg:
            dirs.extend(self.get_xdg_dirs())

        dirs = [os.path.join(d, self.project_name) for d in dirs]

        config_dir_custom = os.environ.get('{}_CONFIG_DIR'.format(self.project_name_upper), None)
        if config_dir_custom:
            dirs.append(config_dir_custom)

        dirs.extend(self.additional_dirs)
        if all:
            dirs.extend(self.additional_hidden_dirs)

        return dirs

    @property
    def installed(self):
        return bool(self.loaded_config_paths)

    def get_config_basename(self):
        """Format the :attr:`config_basename` and return it."""
        basename = self.config_basename
        if basename is None:
            basename = '{}.ini'
        return basename.format(self.project_name)

    def get_config_parser(self):
        return ExtendedConfigParser()

    def get_config_defaults(self):
        return self.config_defaults

    def get_config_paths(self, all=False):
        return [
            os.path.join(d, self.get_config_basename())
            for d in self.get_dirs(all=all)
        ]

    def get_directory_select_class(self):
        return TerminalDirectorySelect

    def init_install(self, path=None):
        select_class = self.get_directory_select_class()
        config_paths = self.get_config_paths(all=True)

        if path:
            path = os.path.normpath(os.path.abspath(path))
            matches = [p for p in config_paths if p.startswith(path)]

            if len(matches) > 1:
                raise InstallationError('More than one possible location in the given path! Please be more specific.')

            if not matches:
                raise InstallationError('Given path does not match any of the available locations!')

            destination = matches[0]
        else:
            select = select_class(self.get_config_paths())
            destination = select.select()

        directory = os.path.dirname(destination)

        try:
            os.makedirs(directory, mode=self.create_mode)
        except FileExistsError:
            pass

        self.install_dir = directory

    def load(self):
        """
        Read .ini files from all possible locations.

        :param require: :class:`bool` - Set this to ``False`` to pass the
            exception.
        :param defaults:
        :return: :class:`configparser.ConfigParser` instance
        """
        config_paths = self.get_config_paths(all=True)
        self.loaded_config_paths = self.config.read(config_paths)
        return self.config

    def get_config(self):
        if not self.installed:
            raise NotInstalled('No configuration file found!')
        return self.config

    def install_file(self, source, basename=None):
        if self.install_dir is None:
            raise InstallationError('Cannot install file outside installation procedure. '
                                    '`install_file` must be called after `init_install`!')
        if basename is None:
            basename = os.path.basename(source)
        destination = os.path.join(self.install_dir, basename)
        shutil.copyfile(source, destination)

    def install_config(self, source):
        basename = self.get_config_basename()
        self.install_file(source, basename)
        if self.set_configdir:
            config_path = os.path.join(self.install_dir, basename)

            with open(config_path, 'r') as f:
                content = f.read()

            subst = '\g<2> = {}'.format(self.install_dir)

            content = re.sub('^(;|#) ?(config ?dir) ?=.*$', subst, content, flags=re.MULTILINE)

            with open(config_path, 'w') as f:
                f.write(content)


class TerminalDirectorySelect(object):
    def __init__(self, path_choices, timeout=None):
        self.path_choices = path_choices
        self.timeout = timeout

    def format_choice(self, index, path):
        if os.path.exists(path):
            path += ' (EXISTS)'
        return '[{}] {}'.format(index, path)

    def select(self):
        """
        Show selection, wait for user input and return it.

        Blocks until valid answer given.

        Returns ``None`` if the timeout is reached.
        """
        print('Select a directory to install the configuration to:')

        for index, path in enumerate(self.path_choices):
            print(self.format_choice(index, path))

        answer = None
        timeout = self.timeout
        while not answer:
            if timeout:
                sys.stdout.write('({}s timeout)'.format(timeout))
            sys.stdout.write(': ')
            sys.stdout.flush()

            i, o, e = pyselect([sys.stdin], [], [], timeout)
            if not i:
                # timeout
                return

            # received something, disable timeout
            timeout = None

            answer = i[0].readline().strip()
            if not answer:
                continue

            try:
                index = int(answer)
            except ValueError:
                sys.stdout.write('"{}" is not a valid number!'.format(answer))
                sys.stdout.flush()
                continue

            try:
                return self.path_choices[index]
            except IndexError:
                sys.stdout.write('"{}" is not a valid choice!'.format(index))
                sys.stdout.flush()
                continue
        return answer


class EditConfigMixin(object):
    def edit(self):
        editor = os.environ.get('EDITOR', None)
        if editor is None:
            print('Environment variable $EDITOR is not set!')
            return 2

        if not hasattr(self, 'get_config_paths'):
            msg = '{} doesn\'t have a `get_config_paths` method! ' \
                  'config_installer.EditConfigMixin must be used with an config_installer.BaseConfigInstaller.'
            raise AssertionError(msg.format(self.__class__.__name__))

        paths = self.get_config_paths(all=True)

        return_code = 3
        for path in paths:
            if not os.path.exists(path):
                continue
            return_code = subprocess.call([editor, path])
            if return_code:
                break

        return return_code
