from __future__ import absolute_import, unicode_literals, print_function

import io
import os
import shutil
import sys
import tempfile
import unittest

from config_installer.installer import BaseConfigInstaller, TerminalDirectorySelect, InstallationError, EditConfigMixin, NotInstalled
from config_installer.parser import ExtendedConfigParser

try:
    from unittest import mock
except ImportError:
    import mock


class FakeConfigInstaller(BaseConfigInstaller):
    project_name = 'test_project'

    enable_etc = True
    enable_xdg = True


class BaseTestCase(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = os.path.join(tempfile.gettempdir(), 'config-installer-tests', 'test-root')

        self.etc_config_dir = os.path.join(self.tmp_dir, 'etc')
        self.etc_config_dir_project = os.path.join(self.etc_config_dir, 'test_project')
        self.etc_config_dir_project_config = os.path.join(self.etc_config_dir_project, 'test_project.ini')

        self.xdg_config_dirs = [os.path.join(self.tmp_dir, 'xdg-config-dir')]

        self.xdg_config_home = os.path.join(self.tmp_dir, 'xdg-config-home')
        self.xdg_config_home_project = os.path.join(self.xdg_config_home, 'test_project')
        self.xdg_config_home_project_config = os.path.join(self.xdg_config_home_project, 'test_project.ini')

        if sys.version_info[0] > 2:
            patcher = mock.patch('builtins.print')
        else:
            patcher = mock.patch('__builtin__.print')
        self.mock_print = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch('config_installer.installer.os.geteuid', return_value=1000)
        self.mock_geteuid = patcher.start()
        self.addCleanup(patcher.stop)

    def tearDown(self):
        shutil.rmtree(self.tmp_dir, ignore_errors=True)


class ConfigInstallerTestCase(BaseTestCase):
    def tearDown(self):
        super(ConfigInstallerTestCase, self).tearDown()
        FakeConfigInstaller.config_defaults = {}

    def test_get_config_parser(self):
        installer = FakeConfigInstaller()

        parser = installer.get_config_parser()

        self.assertIsInstance(parser, ExtendedConfigParser)

    def test_get_config(self):
        os.makedirs(self.xdg_config_home_project)
        with open(self.xdg_config_home_project_config, 'w') as f:
            f.write('[main]\nkey = value\n')
        installer = FakeConfigInstaller()

        config = installer.get_config()

        self.assertIsInstance(config, ExtendedConfigParser)

    def test_get_config_raises_if_not_installed(self):
        installer = FakeConfigInstaller()

        with self.assertRaises(NotInstalled):
            installer.get_config()

    def test_get_directory_select_class(self):
        installer = FakeConfigInstaller()

        cls = installer.get_directory_select_class()

        self.assertEqual(TerminalDirectorySelect, cls)

    def test_get_config_basename(self):
        installer = FakeConfigInstaller()
        installer.config_basename = None

        result = installer.get_config_basename()

        self.assertEqual('test_project.ini', result)

    def test_get_config_basename_custom(self):
        installer = FakeConfigInstaller()
        installer.config_basename = 'config.ini'

        result = installer.get_config_basename()

        self.assertEqual('config.ini', result)

    def test_is_root(self):
        self.mock_geteuid.return_value = 0
        installer = FakeConfigInstaller()

        self.assertTrue(installer.is_root)

    def test_parser_defaults(self):
        FakeConfigInstaller.config_defaults = {
            'key_a': 1,
            'test-key': 'something'
        }
        installer = FakeConfigInstaller()

        self.assertEqual('1', installer.config.get('DEFAULT', 'key_a'))
        self.assertEqual('something', installer.config.get('DEFAULT', 'test-key'))

    def test_parser_defaults_list(self):
        test_list = ['a', 'b', 'something']
        test_list_ini = """
    a,
    b,
    something"""
        FakeConfigInstaller.config_defaults = {
            'test-list': test_list
        }
        installer = FakeConfigInstaller()

        self.assertEqual(test_list_ini, installer.config.get('DEFAULT', 'test-list'))
        self.assertListEqual(test_list, installer.config.getlist('DEFAULT', 'test-list'))


class ConfigInstallerGetXdgDirsTestCase(BaseTestCase):
    def setUp(self):
        super(ConfigInstallerGetXdgDirsTestCase, self).setUp()

        self._old_XDG_CONFIG_DIRS = os.environ.get('XDG_CONFIG_DIRS', None)
        os.environ['XDG_CONFIG_DIRS'] = ':'.join(self.xdg_config_dirs)

        self._old_XDG_CONFIG_HOME = os.environ.get('XDG_CONFIG_HOME', None)
        os.environ['XDG_CONFIG_HOME'] = self.xdg_config_home

        self.installer = FakeConfigInstaller()

    def tearDown(self):
        super(ConfigInstallerGetXdgDirsTestCase, self).tearDown()
        if self._old_XDG_CONFIG_DIRS is not None:
            os.environ['XDG_CONFIG_DIRS'] = self._old_XDG_CONFIG_DIRS
        if self._old_XDG_CONFIG_HOME is not None:
            os.environ['XDG_CONFIG_HOME'] = self._old_XDG_CONFIG_HOME

    def test_user(self):
        result = list(self.installer.get_xdg_dirs())

        self.assertListEqual(
            [
                self.xdg_config_home,
            ],
            result
        )

    def test_root(self):
        self.installer.is_root = True

        result = list(self.installer.get_xdg_dirs())

        self.assertListEqual(
            self.xdg_config_dirs + [self.xdg_config_home],
            result
        )

    def test_root_empty_path_in_env_var(self):
        self.installer.is_root = True
        self.xdg_config_dirs.append(os.path.join(self.tmp_dir, 'xdg-config-dir2'))
        os.environ['XDG_CONFIG_DIRS'] = '{}::{}'.format(*self.xdg_config_dirs)

        result = list(self.installer.get_xdg_dirs())

        self.assertListEqual(
            self.xdg_config_dirs + [self.xdg_config_home],
            result
        )


class ConfigInstallerGetDirsTestCase(BaseTestCase):
    def setUp(self):
        super(ConfigInstallerGetDirsTestCase, self).setUp()

        patcher = mock.patch.object(BaseConfigInstaller, 'get_xdg_dirs', return_value=[self.xdg_config_home])
        self.mock_get_xdg_dirs = patcher.start()
        self.addCleanup(patcher.stop)

        self.installer = FakeConfigInstaller()

    def tearDown(self):
        super(ConfigInstallerGetDirsTestCase, self).tearDown()
        try:
            del os.environ['TEST_PROJECT_CONFIG_DIR']
        except KeyError:
            pass

    def test_user(self):
        result = list(self.installer.get_dirs())

        self.assertListEqual(
            [
                self.xdg_config_home_project,
            ],
            result
        )

    def test_user_without_xdg(self):
        self.installer.enable_xdg = False

        result = list(self.installer.get_dirs())

        self.assertListEqual([], result)

    def test_root(self):
        self.installer.is_root = True

        result = list(self.installer.get_dirs())

        self.assertListEqual(
            [
                os.path.join('/etc', self.installer.project_name),
                self.xdg_config_home_project,
            ],
            result
        )

    def test_root_without_xdg(self):
        self.installer.is_root = True
        self.installer.enable_xdg = False

        result = list(self.installer.get_dirs())

        self.assertListEqual(
            [
                os.path.join('/etc', self.installer.project_name),
            ],
            result
        )

    def test_custom_config_dir(self):
        config_dir = os.path.join(self.tmp_dir, 'test_project_config_dir')
        os.environ['TEST_PROJECT_CONFIG_DIR'] = config_dir

        result = list(self.installer.get_dirs())

        self.assertListEqual(
            [
                self.xdg_config_home_project,
                config_dir
            ],
            result
        )


class MockGetDirsMixin(object):
    def setUp(self):
        super(MockGetDirsMixin, self).setUp()

        patcher = mock.patch.object(BaseConfigInstaller, 'get_dirs',
                                    return_value=[self.xdg_config_home])
        self.mock_get_installation_dirs = patcher.start()
        self.addCleanup(patcher.stop)


class MockGetDirectorySelectClassMixin(object):
    def setUp(self):
        super(MockGetDirectorySelectClassMixin, self).setUp()

        self.mock_directory_select = mock.Mock()
        self.mock_directory_select.select = mock.Mock(return_value=self.xdg_config_home)

        self.select_class = lambda s: self.mock_directory_select
        patcher = mock.patch.object(BaseConfigInstaller, 'get_directory_select_class', return_value=self.select_class)
        self.mock_get_directory_select_class = patcher.start()
        self.addCleanup(patcher.stop)


class BaseConfigInstallerTestCase(MockGetDirsMixin, MockGetDirectorySelectClassMixin, BaseTestCase):
    def setUp(self):
        super(BaseConfigInstallerTestCase, self).setUp()

        self.installer = FakeConfigInstaller()


class ConfigInstallerInstalledTestCase(BaseConfigInstallerTestCase):
    def test_not_installed(self):
        self.installer.loaded_config_paths = []
        self.assertFalse(self.installer.installed)

    def test_installed(self):
        self.installer.loaded_config_paths = []
        self.assertFalse(self.installer.installed)


class ConfigInstallerInstallFileTestCase(BaseConfigInstallerTestCase):
    def setUp(self):
        super(ConfigInstallerInstallFileTestCase, self).setUp()
        os.makedirs(self.xdg_config_home)
        test_file_basename = 'test_file.txt'
        self.test_file = os.path.join(self.tmp_dir, test_file_basename)
        with open(self.test_file, 'w') as f:
            f.write('content')
        self.destination = os.path.join(self.xdg_config_home, test_file_basename)

    def test_raises_if_not_installing(self):
        with self.assertRaises(InstallationError):
            self.installer.install_file(self.test_file)

        self.assertFalse(os.path.exists(self.destination))

    def test(self):
        self.installer.install_dir = self.xdg_config_home

        self.installer.install_file(self.test_file)

        self.assertTrue(os.path.exists(self.destination))

    def test_basename(self):
        self.installer.install_dir = self.xdg_config_home

        self.installer.install_file(self.test_file, 'installed.txt')

        destination = os.path.join(self.xdg_config_home, 'installed.txt')
        self.assertTrue(os.path.exists(destination))
        with open(destination) as f:
            self.assertEqual('content', f.read())


class ConfigInstallerInstallConfigTestCase(BaseConfigInstallerTestCase):
    def setUp(self):
        super(ConfigInstallerInstallConfigTestCase, self).setUp()

        patcher = mock.patch.object(BaseConfigInstaller, 'install_file')
        self.mock_install_file = patcher.start()
        self.addCleanup(patcher.stop)

    def test(self):
        source = os.path.join(self.tmp_dir, 'config.ini')

        self.installer.install_config(source)

        self.mock_install_file.assert_called_once_with(source, 'test_project.ini')


class ConfigInstallerInitInstallTestCase(BaseConfigInstallerTestCase):
    def setUp(self):
        super(ConfigInstallerInitInstallTestCase, self).setUp()

        dirs = [self.etc_config_dir] + self.xdg_config_dirs + [self.xdg_config_home]
        self.mock_get_installation_dirs.return_value = dirs

        # patcher = mock.patch.object(BaseConfigInstaller, 'get_config_paths',
        #                             return_value=[self.xdg_config_home_project_config])
        # self.mock_pyselect = patcher.start()
        # self.addCleanup(patcher.stop)

    def test(self):
        self.installer.init_install()

    def test_path(self):
        self.installer.init_install(self.xdg_config_home)

    def test_path_exists(self):
        os.makedirs(self.xdg_config_home_project)
        self.installer.init_install(self.xdg_config_home)

    def test_path_no_match(self):
        with self.assertRaises(InstallationError):
            self.installer.init_install(os.path.join(self.tmp_dir, 'some_other_path'))

    def test_path_with_multiple_matches(self):
        with self.assertRaises(InstallationError):
            self.installer.init_install(self.tmp_dir)


class MockPySelectMixin(object):
    def setUp(self):
        super(MockPySelectMixin, self).setUp()

        patcher = mock.patch('config_installer.installer.sys')
        self.mock_sys = patcher.start()
        self.addCleanup(patcher.stop)

        self.inputs = [io.BytesIO(b'0')]

        def test_select(r, w, x, timeout=None):
            if not self.inputs:
                raise RuntimeError('No select input!')
            return [self.inputs.pop(0)], [], []

        patcher = mock.patch('config_installer.installer.pyselect', side_effect=test_select)
        self.mock_pyselect = patcher.start()
        self.addCleanup(patcher.stop)


class TerminalDirectorySelectFormatChoiceTestCase(MockPySelectMixin, BaseTestCase):
    def test(self):
        select = TerminalDirectorySelect([])

        result = select.format_choice(0, self.xdg_config_home)

        self.assertEqual('[0] {}'.format(self.xdg_config_home), result)

    def test_adds_exists(self):
        select = TerminalDirectorySelect([])

        os.makedirs(self.xdg_config_home)

        result = select.format_choice(1, self.xdg_config_home)

        self.assertEqual('[1] {} (EXISTS)'.format(self.xdg_config_home), result)


class TerminalDirectorySelectSelectTestCase(MockPySelectMixin, BaseTestCase):
    def setUp(self):
        super(TerminalDirectorySelectSelectTestCase, self).setUp()
        self.select = TerminalDirectorySelect(['a', 'b', 'c'])

    def test_calls_python_select(self):
        result = self.select.select()

        self.mock_pyselect.assert_called_once_with([self.mock_sys.stdin], [], [], None)

    def test_timeout_is_passed_to_python_select(self):
        self.select.timeout = 2

        result = self.select.select()

        self.mock_pyselect.assert_called_once_with([self.mock_sys.stdin], [], [], 2)

    def test_a(self):
        result = self.select.select()

        self.assertEqual('a', result)

    def test_b(self):
        self.inputs = [io.BytesIO(b'1')]

        result = self.select.select()

        self.assertEqual('b', result)

    def test_c(self):
        self.inputs = [io.BytesIO(b'2')]

        result = self.select.select()

        self.assertEqual('c', result)

    def test_not_a_valid_number(self):
        self.inputs = [io.StringIO('a')]

        result = self.select.select()

        self.mock_sys.stdout.write.assert_called_with('"a" is not a valid number!')

    def test_not_a_valid_choice(self):
        self.inputs = [io.BytesIO(b'99')]

        result = self.select.select()

        self.mock_sys.stdout.write.assert_called_with('"99" is not a valid choice!')

    def test_timeout(self):
        self.mock_pyselect.side_effect = None
        self.mock_pyselect.return_value = [], [], []

        result = self.select.select()

        self.assertIsNone(result)

    def test_empty_answer(self):
        self.inputs = [io.BytesIO(b''), io.BytesIO(b'1')]

        result = self.select.select()

        self.assertEqual('b', result)


class FakeEditConfigInstaller(EditConfigMixin, FakeConfigInstaller):
    pass


class EditConfigTestCase(BaseTestCase):
    def setUp(self):
        super(EditConfigTestCase, self).setUp()

        patcher = mock.patch.object(FakeEditConfigInstaller, 'get_config_paths',
                                    return_value=[self.xdg_config_home_project_config])
        self.mock_subprocess_call = patcher.start()
        self.addCleanup(patcher.stop)

        self._old_EDITOR = os.environ.get('EDITOR', None)
        self.editor = 'my-awesome-config-installer-test-editor'
        os.environ['EDITOR'] = self.editor

        os.makedirs(self.xdg_config_home_project)
        with open(self.xdg_config_home_project_config, 'w') as f:
            f.write('[section]\nkey = value')

        self.installer = FakeEditConfigInstaller()

        patcher = mock.patch('config_installer.installer.subprocess.call', return_value=0)
        self.mock_subprocess_call = patcher.start()
        self.addCleanup(patcher.stop)

    def tearDown(self):
        super(EditConfigTestCase, self).tearDown()
        if self._old_EDITOR is not None:
            os.environ['EDITOR'] = self._old_EDITOR

    def test(self):
        returncode = self.installer.edit()

        self.mock_subprocess_call.assert_called_once_with([self.editor, self.xdg_config_home_project_config])
        self.assertEqual(0, returncode)

    def test_no_editor(self):
        try:
            del os.environ['EDITOR']
        except KeyError:
            pass

        returncode = self.installer.edit()

        self.mock_print.assert_called_once_with('Environment variable $EDITOR is not set!')
        self.assertEqual(0, self.mock_subprocess_call.call_count)
        self.assertEqual(2, returncode)

    def test_editor_error(self):
        self.mock_subprocess_call.return_value = 5

        returncode = self.installer.edit()

        self.assertEqual(1, self.mock_subprocess_call.call_count)
        self.assertEqual(5, returncode)

    def test_invalid_mixin(self):
        with self.assertRaises(AssertionError):
            EditConfigMixin().edit()

    def test_config_is_skipped_if_it_doesnt_exist(self):
        os.unlink(self.xdg_config_home_project_config)

        returncode = self.installer.edit()

        self.assertEqual(0, self.mock_subprocess_call.call_count)
        self.assertEqual(3, returncode)

