from __future__ import absolute_import, unicode_literals, print_function

import io
import unittest

from config_installer.parser import ExtendedConfigParser

try:
    from configparser import NoOptionError, NoSectionError
except ImportError:
    from ConfigParser import NoOptionError, NoSectionError

try:
    from unittest import mock
except ImportError:
    import mock


TEST_CONFIG = """\
[get]
key = value

[getboolean]
t1 = True
t2 = true
t3 = yes
f1 = False
f2 = false
f3 = no
E = invalid

[getint]
a = 1
b = -1
E = invalid

[getlist]
a = 1, 2, 3
b = 1
    2
    3
c = 1,
    2,
    3
d = 1,
    2,
    3,
e = 1 , 2 ,3
f = 1
    2  ,
    3,
z = 1
"""


class BaseTestCase(unittest.TestCase):
    def setUp(self):
        self.parser = ExtendedConfigParser()
        self.parser.readfp(io.StringIO(TEST_CONFIG))


class ExtendedConfigParserGetTestCase(BaseTestCase):
    def test(self):
        self.assertEqual('value', self.parser.get('get', 'key'))

    def test_fallback(self):
        self.assertEqual('123', self.parser.get('get', 'unknown', fallback='123'))


class ExtendedConfigParserGetBooleanTestCase(BaseTestCase):
    def test_true(self):
        self.assertTrue(self.parser.getboolean('getboolean', 't1'))
        self.assertTrue(self.parser.getboolean('getboolean', 't2'))
        self.assertTrue(self.parser.getboolean('getboolean', 't3'))

    def test_false(self):
        self.assertFalse(self.parser.getboolean('getboolean', 'f1'))
        self.assertFalse(self.parser.getboolean('getboolean', 'f2'))
        self.assertFalse(self.parser.getboolean('getboolean', 'f3'))

    def test_invalid(self):
        with self.assertRaises(ValueError):
            self.parser.getboolean('getboolean', 'E')

    def test_unknown_option(self):
        with self.assertRaises(NoOptionError):
            self.assertTrue(self.parser.getboolean('getboolean', 'unknown'))

    def test_fallback(self):
        self.assertTrue(self.parser.getboolean('getboolean', 'unknown', fallback=True))

    def test_fallback_different_type(self):
        self.assertIsNone(self.parser.getboolean('getboolean', 'unknown', fallback=None))


class ExtendedConfigParserGetIntTestCase(BaseTestCase):
    def test(self):
        self.assertEqual(1, self.parser.getint('getint', 'a'))

    def test_negative(self):
        self.assertEqual(-1, self.parser.getint('getint', 'b'))

    def test_invalid(self):
        with self.assertRaises(ValueError):
            self.parser.getint('getint', 'E')

    def test_unknown_option(self):
        with self.assertRaises(NoOptionError):
            self.assertTrue(self.parser.getint('getint', 'unknown'))

    def test_fallback(self):
        self.assertEqual(5, self.parser.getint('getint', 'unknown', fallback=5))


class ExtendedConfigParserGetListTestCase(BaseTestCase):
    def test(self):
        for key in ['a', 'b', 'c', 'd', 'e', 'f']:
            self.assertListEqual(
                ['1', '2', '3'],
                self.parser.getlist('getlist', key),
                msg='Key {}'.format(key)
            )

    def test_single_element(self):
        self.assertListEqual(['1'], self.parser.getlist('getlist', 'z'))

    def test_unknown_option(self):
        with self.assertRaises(NoOptionError):
            self.assertTrue(self.parser.getlist('getlist', 'unknown'))

    def test_fallback(self):
        self.assertListEqual(
            ['a', 'b'],
            self.parser.getlist('getlist', 'unknown', fallback=['a', 'b'])
        )
